/****************************************************************
 * Skyler Schneider ss868                                       *
 * Jack Ellsworth   jfe5                                        *
 * ECE 4760 Final Project                                       *
 * USB Wireless Tilt Mouse                                      *
 ****************************************************************/

// Transmitter unit consists of two accelerometers and wireless TX.

/****************************************************************
 * Hardware Connections                                         *
 ****************************************************************/

/*
 * PORTA
 *  A.0 - ADC Channel 0 (y Accelerometer)
 *  A.1 - ADC Channel 1 (x Accelerometer)
 */

/*
 * PORTC
 *  C.0 - Right  Pushbutton
 *  C.1 - Middle Pushbutton
 *  C.2 - Top    Pushbutton (Calibrate)
 *  C.3 - Left   Pushbutton
 */

/*
 * PORTD
 *  D.1 - Wireless transmitter
 */

/****************************************************************
 * Compiler Directives                                          *
 ****************************************************************/

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "txrx_usart0_head.c"

// Wireless definitions
#define data_length  8
#define tx_id        3
#define threshold    17

/****************************************************************
 * Global Declarations                                          *
 ****************************************************************/

// Structure for storing mouse status data to be sent wirelessly
typedef struct {
    uint8_t buttonMask;
    int8_t  dx;
    int8_t  dy;
    int8_t  dWheel;
} report_t;

report_t reportBuffer;

// Wireless variables
char     data[data_length]; 
uint8_t  packet_count;
uint16_t time;

// ADC digital output
uint8_t  Ain;
uint8_t  measured;
uint8_t  measureMode; // 0 for y accelerometer, 1 for x accelerometer

// Timer1 variables for 16.384ms period
uint16_t TCNT1_prev;
uint16_t TCNT1_current;
uint16_t TCNT1_change;

// Acceleration variables
uint8_t Athreshx;
uint8_t Athreshy;
int8_t  ax;
int8_t  ay;

// Switch reading variables
uint8_t SW;
uint8_t SW_prev;
uint8_t SW_debounced;
uint8_t SW_debounced_prev;
uint8_t SW_pressed;
uint8_t calibrate;
uint8_t leftHeld;
uint8_t middleHeld;
uint8_t rightHeld;

void init(void); // initialization method

/****************************************************************
 * Interrupt Service Routines                                   *
 ****************************************************************/

ISR (TIMER0_COMPA_vect) {
    if(time <= threshold) time++;
}

/****************************************************************
 * Helper Functions                                             *
 ****************************************************************/

void generate_data(char count) {
    data[0] = 0x06;                       // IDENTIFIER BYTE
    data[1] = reportBuffer.buttonMask;    // BUTTON BYTE
    data[2] = reportBuffer.dx & 0x0F;     // X BYTE L
    data[3] = reportBuffer.dx >> 4;       // X BYTE H
    data[4] = reportBuffer.dy & 0x0F;     // Y BYTE L
    data[5] = reportBuffer.dy >> 4;       // Y BYTE H
    data[6] = reportBuffer.dWheel & 0x0F; // WHEEL BYTE L
    data[7] = reportBuffer.dWheel >> 4;   // WHEEL BYTE H
}

// Return 1 if SW has transitioned from unpressed to pressed
uint8_t keypress_SW() {
    SW_prev = SW;
    SW_debounced_prev = SW_debounced;
    // Read SW
    SW = (~PINC & 0x0F) ^ 0b00000110; // hack - middle and calibrate buttons are reverse operation
    // Debounce SW
    if(SW == SW_prev) SW_debounced = SW;
    // Detect transition from unpressed to pressed
    return (SW_debounced) && (!SW_debounced_prev);
}

/****************************************************************
 * Main Routine                                                 *
 ****************************************************************/
int main(void) {
    int8_t tempInt;
  	uint8_t switcha = 0x00;  
    init();
  
    while(1) {
        // Send wireless information every 17ms
        if(time == threshold) {
            // Executes every threshold ms
            //generate_data(packet_count);
			data[0] = (switcha) ? 0 : 1;
			switcha ^= 0x01;
            tx_me(data, data_length, tx_id);
            time = 0;
            packet_count++;
        }
        
        // Probe an accelerometer every 16.384ms
        TCNT1_current = TCNT1;
        TCNT1_change  = TCNT1_current - TCNT1_prev;
        if(TCNT1_change > 255) {
            TCNT1_prev = TCNT1_current;
            // Get ADC sample
            Ain = ADCH;
            // Switch input to alternate accelerometer reads
            ADMUX = ADMUX ^ 0x01;
            // Start another conversion
            ADCSRA |= (1<<ADSC);
            // Signal measurement
            measured = 1;
            PORTD = PORTD ^ 0b00000100;
        }
        
        // Read pushbuttons at same time as accelerometers
        if(measured) {
            SW_pressed = keypress_SW();
            calibrate  = (SW_debounced & 0b00000100) >> 2;
            leftHeld   = (SW_debounced & 0b00001000) >> 3;
            middleHeld = (SW_debounced & 0b00000010) >> 1;
            rightHeld  = (SW_debounced & 0b00000001);
            
            
            if(measureMode == 0) {
                // y accelerometer
                if(calibrate) Athreshy = Ain;
                tempInt = (((int16_t)Ain) - ((int16_t)Athreshy)) >> 1;
                ay = (ay >> 1) + (tempInt >> 1);
            }
            else {
                // x accelerometer
                if(calibrate) Athreshx = Ain;
                tempInt = (((int16_t)Ain) - ((int16_t)Athreshx)) >> 1;
                ax = (ax >> 1) + (tempInt >> 1);
            }
            measureMode = measureMode ^ 0x01;
            measured = 0;
            
            // Update report buffer
            reportBuffer.dx = ax;
            reportBuffer.dy = ay;
            reportBuffer.dWheel = 0;
            reportBuffer.buttonMask = (middleHeld << 2) | (rightHeld << 1) | leftHeld;
        }
    }
}

/****************************************************************
 * Initialization Routine                                       *
 ****************************************************************/

void init() {
    
    // Set up ports
    DDRA = 0x00; // ADC inputs
    DDRC = 0x00; // pushbutton inputs
    DDRD = 0xFF; // wireless tx output
    
    // Set up Timer0
    TIMSK0 = (1<<OCIE0A); // Turn on Timer0 comp match ISR
    OCR0A  = 250;         // Set the compare reg to 250 time ticks (1ms)
    // Prescalar to 64 and turn on clear-on-match
    TCCR0A = (1<<WGM01);  // Turn on clear-on-match
    TCCR0B = 0x03;        // Clock prescalar to 64
    
    // Set up wireless transmitter
    txrx_init(1,0,249,0); // TX only - 4000 baud - led off
    
    // Set up Timer1
    TCCR1B = 0x05; // Timer1 runs at prescalar 1024
    TCNT1_current = TCNT1;
    TCNT1_prev    = TCNT1_current;
    TCNT1_change  = 0;
    
    // Init the A to D converter
    // Left adj - digital result stored in {ADCH,ADCL} as {0bxxxxxxxx_xx000000}
    // Aref = AVcc
    // Channel 0 (A.0) implied by not setting any channel select bits
    ADMUX = (1 << REFS0) | (1<<ADLAR);
    
    // Enable ADC and set prescaler to 1/128*16MHz = 125,000Hz
    // 8us/clk -> 0.1ms per conversion
    // Clear interupt enable
    // Start a conversion
    ADCSRA = ((1<<ADEN) | (1<<ADSC)) + 7;
    
    // Initialize variables
    Ain         = 0;
    measured    = 0;
    measureMode = 0;
    Athreshx    = 131;
    Athreshy    = 129;
    ax          = 0;
    ay          = 0;
    
    SW                = 0;
    SW_prev           = 0;
    SW_debounced      = 0;
    SW_debounced_prev = 0;
    SW_pressed        = 0;
    calibrate         = 0;
    leftHeld          = 0;
    middleHeld        = 0;
    rightHeld         = 0;
    
    reportBuffer.buttonMask = 0;
    reportBuffer.dx         = 0;
    reportBuffer.dy         = 0;
    reportBuffer.dWheel     = 0;
    
    time = 0;
    packet_count = 0;
    generate_data(packet_count);
    
    // Enable interrupts
    sei();
}
